# -*- coding: utf-8 -*-

from kronbinations.kronbinations import *
from kronbinations.JIT_kronbinations import *
from kronbinations.JIT_kronbinations_load import *
from kronbinations.JIT_Array import *
from kronbinations.Kron_Array import *
from kronbinations.Kron_Fun_Modifier import *


import pytest
import numpy as np
from kronbinations import JIT_kronbinations

# Generate lists of different dtypes
# and test the functions on them
floats = np.array([1.2, 3.4, 5.6, 7.8, 9.0])
ints = np.array([1, 3, 5, 7, 9], dtype=int)
strings = np.array(['a', 'b', 'c', 'd', 'e'])

# Create a kronbinations object
def test_kronbinations_matrices():
    # Create a kronbinations object
    # with the lists as values
    def fun(k, A, B):
        for i, v in k.kronprod(change=False, redo=False):
            A[i] = v[0]+v[1]+i[2]
            B[i] = v[0]-v[1]
        return A, B
    k = JIT_kronbinations(floats, ints, strings, func=fun, redo=False)
    assert k.size == 125
    assert k.shape == (5,5,5)
    assert k.ndim == 3
    # Test the generation of matrices
    o = k.ones(dtype=int)
    z = k.zeros(dtype=int)
    e = k.empty()
    f = k.full(fill_value= 1.0/3.0)
    assert o.shape == (5,5,5)
    assert z.shape == (5,5,5)
    assert e.shape == (5,5,5)
    assert f.shape == (5,5,5)
    
def test_kronbinations_loop_outputs():
    b = [False, True]
    # Check number of outputs
    for index in b:
        for change in b:
            how_many = 1 if index else 0
            how_many += 1 if change else 0
            if how_many == 0:
                how_many = 2 # elements of the values tuple - 1
                def funny(k, A):
                    for output in k.kronprod(index=index, change=change, progress=False):
                        assert len(output) == how_many+1
                k = JIT_kronbinations(floats, ints, strings, func=funny, redo=False, autosave=False)
                A = k.zeros()
                k.calculate_all()
                
def test_krombinations_intermediate_outputs():
    # Check the intermediate outputs
    def funny(k, A):
        for i,v,c in k.kronprod(index=True, change=True, progress=False):
            assert k.changed() == c
            assert k.index() == i # works since switching to tuples
            assert k.value() == v
            for j in range(3):
                assert k.changed(j) == c[j]
                assert k.index(j) == i[j]
                assert k.value(j) == v[j]
    k = JIT_kronbinations(floats, ints, strings, func=funny, redo=False, autosave=False)
    k.zeros()
    k.calculate_all()

def test_krombinations_settings_and_changes_to_them():
    def fun(k, A, B):
        for i, v in k.kronprod(change=False):
            A[i] = v[0]+v[1]+i[2]
            B[i] = v[0]-v[1]
        return A, B
    k = JIT_kronbinations(floats, ints, strings, func=fun, redo=False, autosave=False)
    k.set(do_index=True, do_change=True, do_tqdm=True)
    do_index, do_change, do_tqdm, return_as_dict = k.get('do_index', 'do_change', 'do_tqdm', 'return_as_dict')
    assert do_index
    assert do_change
    assert do_tqdm
    k.set(do_index=False, do_change=False, do_tqdm=False)
    do_index, do_change, do_tqdm, return_as_dict = k.get('index', 'do_change', 'do_tqdm', 'return_as_dict')
    assert not do_index
    assert not do_change
    assert not do_tqdm
    
def test_kronbinations_illegal_dict_outputs():
    # Check the illegal dictionary outputs
    def funny(k, A):
        for i,v,c in k.kronprod(index=True, change=True, progress=False):
            k.index('floats')
    k = JIT_kronbinations(floats, ints, strings, func=funny, redo=False, autosave=False)
    k.empty()
    
    def funny2(k, A):
        for i,v,c in k.kronprod(index=True, change=True, progress=False):
            k.value('floats')
    k2 = JIT_kronbinations(floats, ints, strings, func=funny2, redo=False, autosave=False)
    k2.empty()
    
    def funny3(k, A):
        for i,v,c in k.kronprod(index=True, change=True, progress=False):
            k.changed('floats')
    k3 = JIT_kronbinations(floats, ints, strings, func=funny3, redo=False, autosave=False)
    k3.empty()
    
    with pytest.raises(ValueError):
        k.calculate_all()
    with pytest.raises(ValueError):
        k2.calculate_all()
    with pytest.raises(ValueError):
        k3.calculate_all()

def test_kronbinations_dict_objects():    
    # Create a kronbinations object
    # with the lists as a dictionary
    keys = ['floats', 'ints', 'strings']
    d = {'floats': floats, 'ints': ints, 'strings': strings}
    def funny(k, A):
        for i,v,c in k.kronprod(index=True, change=True, progress=False):
            for j, key in enumerate(keys):
                assert k.changed(key) == c[key]
                assert k.index(key) == i[j]
                assert k.value(key) == v[key]
    with pytest.raises(ValueError):
        k = JIT_kronbinations(d, d, func=funny, redo=False, autosave=False)
    k = JIT_kronbinations(d, func=funny, redo=False, autosave=False)
    k.empty()
    k.calculate_all()


# -*- coding: utf-8 -*-
import numpy as np
import itertools
import warnings
import os


class JIT_Array(np.ndarray):
    def __new__(cls, parent, type, *args, redo=False, name=None, **kwargs):
        data_dir = parent.data_dir
        checksum = parent.checksum
        if name is None:
            name = 'array_' + str(parent.how_many_arrays_set)
        filename = os.path.join(data_dir, checksum + '_' + name + '.npz')
        file_exists = os.path.isfile(filename)
        if file_exists and not redo:
            try:
                with np.load(filename, allow_pickle=True) as f:
                    var = f['var']
                    calculated = f['calculated']
                    how_many_missing = f['how_many_missing']
            except:
                raise Exception('Error loading file: ' + filename)
            obj = var.view(cls)
            obj.calculated = calculated
            obj.how_many_missing = how_many_missing
            shape_parent = parent.shape
            if len(obj.shape) > len(shape_parent):
                added_shape = len(obj.shape) - len(shape_parent)
                shape_added = list(obj.shape[-added_shape:])
            else:
                added_shape = 0
                shape_added = []
        else:
            # first args is an addition to shape, so we need to add it to the shape of the parent 
            if len(args) > 0: #parent.shape is a tuple
                if isinstance(args[0], tuple):
                    shape_added = list(args[0])
                elif isinstance(args[0], int):
                    shape_added = [args[0]]
                elif isinstance(args[0], list):
                    shape_added = args[0]
                else:
                    shape_added = []
                shape_parent = parent.shape
                shape = tuple(list(shape_parent) +shape_added)
                if len(args) > 1:
                    args = args[1:]
                else:
                    args = ()
                added_shape = len(shape_added)
            else:
                shape = parent.shape
                shape_parent = shape
                added_shape = 0
            if type=='ones':
                obj = np.ones(shape, *args, **kwargs).view(cls)
            elif type=='zeros':
                obj = np.zeros(shape, *args, **kwargs).view(cls)
            elif type=='full':
                obj = np.full(shape, *args, **kwargs).view(cls)
            else:
                obj = np.empty(shape, *args, **kwargs).view(cls)
            obj.calculated = np.zeros(shape_parent, dtype=bool)
            obj.how_many_missing = np.prod(shape_parent)  
        obj.added_shape = added_shape # did we add to the shape of the parent
        obj.shape_parent = shape_parent
        obj.ndim_parent = len(shape_parent)
        obj.parent = parent
        obj.name = name
        obj.checksum = checksum
        obj.data_dir = data_dir
        obj.filename = filename
        obj.file_exists = file_exists
        return obj
    
    # A function to save the matrix
    def save(self):
        # if directory doesn't exist, create it
        filename = self.filename
        if not os.path.exists(self.data_dir):
            try:
                os.makedirs(self.data_dir)
            except:
                raise Exception('Error: Creating directory. ' +  self.data_dir)
        try:
            # ignore Warnings for saving
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                np.savez(filename, var=np.asarray(self), calculated=np.asarray(self.calculated), how_many_missing=self.how_many_missing, allow_pickle=True)
        except:
            raise Exception('Error: Saving file. ' +  filename)
        self.file_exists = True

    def slice_to_tuple(self, s, array_shape):
        if isinstance(s, slice):
            return np.array(range(*s.indices(array_shape)))
        else:
            # return as iterable
            if isinstance(s, (int, np.int32, np.int64)):
                return np.array([s])
            elif isinstance(s, list):
                return np.array(s)
            else:
                return s
    # function that transforms a tuple of slices into index combinations
    # Construct all combinations of indexes from slices
    def slice_tuple_to_tuple(self, slice_tuple, array_shape):
        # check for ellipsis and add slices if needed
        if slice_tuple[0] is Ellipsis:
            # remove ellipsis and add slices to the beginning of the tuple
            slice_tuple = (slice(None, None, None),)*(len(array_shape)-len(slice_tuple)+1) + slice_tuple[1:]
        elif slice_tuple[-1] is Ellipsis:
            # remove ellipsis and add slices to the end of the tuple
            slice_tuple = slice_tuple[:-1] + (slice(None, None, None),)*(len(array_shape)-len(slice_tuple)+1)
        else:
            if len(slice_tuple) < len(array_shape):
                slice_tuple = slice_tuple + (slice(None, None, None),)*(len(array_shape)-len(slice_tuple))
        return [self.slice_to_tuple(s, a) for s, a in zip(slice_tuple, array_shape)]

    def find_element_indexes(self, slice_tuple, array_shape, indexes):
        tuples = self.slice_tuple_to_tuple(slice_tuple, array_shape)
        lengths = [1]
        lengths += [len(t) for t in np.flip(tuples[1:])]
        len_prod = np.flip(np.cumprod(lengths))
        len_tuples = len(tuples)
        # find indexes in the tuples using mod operations
        inds = np.zeros([len(indexes), len_tuples], dtype=int)
        for i in range(len_tuples):
            inds[:,i] = tuples[i][ indexes // len_prod[i] ]
            indexes = indexes % len_prod[i]
        return inds
    
    def find_all_elements(self, slice_tuple, array_shape):
        tuples = self.slice_tuple_to_tuple(slice_tuple, array_shape)
        lengths = [1]
        lengths += [len(t) for t in np.flip(tuples[1:])]
        how_many = np.prod(lengths)
        inds = np.zeros([how_many, len(lengths)], dtype=int)
        # use itertools combinations to find all combinations of indexes
        for i, t in enumerate(itertools.product(*tuples)):
            inds[i,:] = t
        return inds
    
    def arg_index_formatting(self, arg):
        #print('arg_index_formatting', arg)
        if self.added_shape > 0: # check if tuple with first element tuple
            if isinstance(arg, tuple) and isinstance(arg[0], tuple):
                arg = arg[0] + arg[1:]
            #if isinstance(arg, tuple) and isinstance(arg[0], np.ndarray): # add arg1 to arg0 copy elements of arg1 to arg0
            #    # transform ndarray of arg[0] into tuple and append arg[1]
            #    arg = tuple(arg[0].flatten()) + arg[1:]
        find_indexes = True
        if not isinstance(arg, tuple):
            if isinstance(arg, int):
                arg = (arg,)
            elif isinstance(arg, list):
                arg = np.array(arg)
                find_indexes = False
            elif isinstance(arg, np.ndarray):
                find_indexes = False
            if arg is Ellipsis:
                arg = (arg,)
                find_indexes = True
            elif isinstance(arg, slice):
                arg = (arg,)
                find_indexes = True
            else:
               arg = (arg,)
        return arg, find_indexes

    def __getitem__(self, arg):
        # add zero to *args to make sure that the first index is not a tuple
        arg, find_indexes = self.arg_index_formatting(arg)
        arg_shortened = arg[:self.ndim_parent]
        if self.how_many_missing > 0:
            items = self.calculated.__getitem__(arg_shortened).flatten() # 
            if not np.all(items):
                # Transform 
                # find the indices that are not yet calculated and calculate them, outputs a list of index combinations
                calc_indexes = np.where(~items)[0]
                if find_indexes == True:
                    indexes = self.find_element_indexes(arg_shortened, self.shape_parent, calc_indexes)
                else:
                    indexes = arg_shortened
                self.parent.calculate(indexes)
        return super().__getitem__(arg).view(np.ndarray)
    
    def is_done(self, index):
        return self.calculated[index]

    def __setitem__(self, arg, value):
        # add zero to *args to make sure that the first index is not a tuple
        arg, _ = self.arg_index_formatting(arg)
        arg_shortened = arg[:self.ndim_parent]
        # get the corresponding calculated items
        if self.how_many_missing > 0:
            number_of_not_yet_set = np.sum(self.calculated[arg_shortened].flatten() == False)
            self.calculated.__setitem__(arg_shortened, True)
            self.how_many_missing -= number_of_not_yet_set
        # set the values
        super().__setitem__(arg, value)

    #def names2dirs(self, *names):
    #    # finds the directions of the hyperarray from the names, then returns the directions

    def asarray(self):
        if self.how_many_missing > 0:
            self.parent.calculate_all()
        return np.asarray(self)
    
    def array(self):
        return self.asarray()
    
    #def dirs_array(self, *names):
    #    # picks out the directions from the array specified by names, 


# -*- coding: utf-8 -*-
import numpy as np
from weighted_tqdm import weighted_kronbinations_tqdm, weighted_tqdm

import inspect
from hashlib import sha1
import os
import numpy as np
import inspect

# import JIT_Array and Kron_Fun_Modifier    
from .JIT_Array import *
from .Kron_Fun_Modifier import *

# An array class, that stores an array, and whether it's values have been calculated,
# and if so, what the values are
# If values of the array are requested, and they have not been calculated, they are calculated using a function handle passed to the constructor
class JIT_kronbinations():
    def __init__(self, *values, func=None, other_func=[], import_statements=[], other_arguments=[], checksum=None, checksum_pre='', autosave=True, data_dir='Cache', redo=False, progress=True, **kwargs):
        # Calculate checksums
        weights_given = True if 'weights' in kwargs.keys() else False
        if not os.path.exists(data_dir):
            os.makedirs(data_dir)
        self.data_dir = data_dir
        if checksum is None:
            self.checksum = self.make_checksum(checksum_pre, func, *values, *import_statements, *other_arguments)
        else:
            self.checksum = checksum
        # check if data_dir exists
        
        self.autosave = autosave
        self.func = func
        self.other_func = other_func
        self.how_many_arrays_set = 0
        self.JIT_Arrays = []
        self.import_statements = import_statements#
        if isinstance(other_arguments, dict):
                self.other_arguments = [other_arguments]
        else:
            self.other_arguments = other_arguments

        # lengths of the values -> ignore the length one arrays, they are not to be iterated over
        # if values is a dictionary, then the keys are the directions, and the values are the arrays
        if isinstance(values[0], dict):
            if len(values) > 1:
                raise ValueError('If values is a dictionary, it must be the only argument')
            input_values = values[0]
            keys = input_values.keys()
            values = input_values.values()
            self.array_vars_all_names, self.array_vars_all, self.all_weights = [], [], []
            for i, (key, value) in enumerate(zip(keys, values)):  
                # check if value is a dict with value and weight
                self.array_vars_all_names.append(key)
                if isinstance(value, dict):
                    if 'value' in value.keys():
                        self.array_vars_all.append(value['value'])
                    else:
                        raise ValueError('If value is a dictionary, it must have a value key') 
                    if 'weight' in value.keys():
                        self.all_weights.append(value['weight'])
                    else:
                        self.all_weights.append(None)
                else: # Normal 
                    self.array_vars_all.append(value)
                    if weights_given:
                        self.all_weights.append(kwargs['weights'][i])
                    else:
                        self.all_weights.append(None)
            self.return_as_dict = True
        else:
            self.array_vars_all = list(values)
            if weights_given:
                self.all_weights = list(kwargs['weights'])
            else:
                self.all_weights = [None] * len(self.array_vars_all)
            self.return_as_dict = False
            self.array_vars_all_names = None
        for i, arr in enumerate(self.array_vars_all):
            # if does not have length, transform into array
            if not hasattr(arr, '__len__'):
                self.array_vars_all[i] = np.array([arr])
        # only relevant values in array_directions
        self.array_vars = [arr for arr in self.array_vars_all if len(arr) > 1]
        if isinstance(values, dict):
            self.array_vars_names = [name for name, arr in zip(self.array_vars_all_names, self.array_vars_all) if len(arr) > 1]
        # add index values of the array vars 
        self.array_vars_indexes = [i for i, arr in enumerate(self.array_vars_all) if len(arr) > 1] 
        self.weights = [w for i, w in enumerate(self.all_weights) if i in self.array_vars_indexes]
        
        if self.return_as_dict:
            self.curr_vals = {key: arr[0] for key, arr in zip(self.array_vars_all_names, self.array_vars_all)}
        else:
            self.curr_vals = [arr[0] for arr in self.array_vars_all]

        self.array_lengths_all = [len(arr) for arr in self.array_vars_all]
        self.array_lengths = [len(arr) for arr in self.array_vars]
        self.shape_all = tuple(self.array_lengths_all)
        self.shape = tuple(self.array_lengths)
        self.ndim_all = len(self.array_lengths_all)
        self.ndim = len(self.array_lengths)
        self.size_all = np.prod(self.array_lengths_all)
        self.size = np.prod(self.array_lengths)

        self.do_index = True
        self.do_change = True
        self.do_tqdm = progress
        self.redo = redo
        self.set(**kwargs)   # redo these values if passed as kwargs
        self.pbar = weighted_kronbinations_tqdm(self.array_vars, self.weights, self.size)
        
        self.curr_index = -1
        self.func_modifier()

    def set(self, **args):
        key_substitution_list = [['index', 'do_index'], ['change', 'do_change'], ['progress', 'do_tqdm']]
        key_list = [v[0] for v in key_substitution_list]
        subs_list = [v[1] for v in key_substitution_list]
        for key, value in args.items():
            # Substitute certain keys from substitution list
            if key in key_list:
                key = subs_list[key_list.index(key)]
            if (key == 'return_as_dict' and value==True) and not isinstance(self.array_vars_all_names, list):
                raise ValueError('Keys are not defined, must create Object via dictionary in order to set "return_as_dict = True".')
            else:
                setattr(self, key, value)
        if self.do_tqdm:
            self.pbar = weighted_kronbinations_tqdm(self.array_vars, self.weights, self.size)
    def get(self, *args):
        key_substitution_list = [['index', 'do_index'], ['change', 'do_change'], ['progress', 'do_tqdm']]
        key_list = [v[0] for v in key_substitution_list]
        subs_list = [v[1] for v in key_substitution_list]
        x = []
        for key in args:
            if key in key_list:
                key = subs_list[key_list.index(key)]
            x.append(getattr(self, key))
        if len(x) == 1:
            return x[0]
        else:
            return x

    def change_all_dtype(self, object, dtype='int'):
        # navigates through the objects lists tuples and dictionaries until it finds a numpy array 
        # and changes the dtype if it is an integer to 'int'
        if isinstance(object, (list, tuple, dict)):
            # recursively call the function on the elements of the object
            is_tuple = isinstance(object, tuple)
            if is_tuple:
                object = list(object)
            if isinstance(object, dict):
                # navigate every key value pair
                for key, value in object.items():
                    object[key] = self.change_all_dtype(value, dtype=dtype)
            else: # isinstance(object, list):
                for i in range(len(object)):
                    object[i] = self.change_all_dtype(object[i], dtype=dtype)
            if is_tuple:
                object = tuple(object)
        elif isinstance(object, np.ndarray):
            if 'int' in str(object.dtype):
                object = object.astype(dtype)
        return object
    
    def make_checksum(self, checksum_pre, func, *args):
        # check every arg in args, if arg is an object of a class 
        # change all dtype int
        args = self.change_all_dtype(args)
        checksum_sha1 = sha1(str(args).encode('utf-8')).hexdigest()
        checksum_sin_fun = checksum_pre + checksum_sha1
        # get function name from func
        if func is not None:
            func_name = func.__name__
            checksum = checksum_pre + checksum_sha1 + func_name
            # check if files with checksum exist
            checksum_exists = False
            dir_list = os.listdir(self.data_dir)
            for file in dir_list:
                if checksum in file:
                    checksum_exists = True
                    break
            if not checksum_exists:
                # check if files with checksum_sin_fun exists -> rename to checksum
                for file in dir_list:
                    if checksum_sin_fun in file:
                        new_filename = file.replace(checksum_sin_fun, checksum)
                        os.rename(os.path.join(self.data_dir, file), os.path.join(self.data_dir, new_filename))  
                        print('  -> Renamed file from ' + file + ' to ' + new_filename)
        else:
            checksum = checksum_sin_fun
        return checksum

    def __getitem__(self, key):
        # If the key is not in the data, return None
        if key in key_list:
            key = subs_list[key_list.index(key)]
        if key not in self.data:
            return None
        else:
            return self.data[key]

    def save(self):
        # Save the data onto every JIT_Array
        for arr in self.JIT_Arrays:
            arr.save()

    # Replace these by LazyArrays and increase self.how_many_arrays_set by one every time
    def empty(self, *var, **args):
        return self.make_array('empty', *var, redo=self.redo, **args)
    def ones(self, *var, **args):
        return self.make_array('ones', *var, redo=self.redo, **args)
    def zeros(self, *var, **args):
        return self.make_array('zeros', *var, redo=self.redo, **args)
    def full(self, *var, **args):
        return self.make_array('full', *var, redo=self.redo, **args)

    def make_array(self, type, *var, redo=False, **args):
        # if name in **args, then use that as the name of the array and remove it from **args
        if 'name' in args:
            name = args['name']
            del args['name']
        else:
            name = None
        new_array = JIT_Array(self, type, *var, **args, name=name, redo=redo)
        self.JIT_Arrays.append(new_array)
        self.how_many_arrays_set += 1
        return new_array

    # This can be impolemented more eleganty by using a kronbinations like approach and defining func later
    def calculate(self, indexes):
        # indexes is a 2d array, where the first dimension is the index of the array, and the second dimension is the index of the value
        n = len(indexes)
        self.setup_iterator(indexes)
        # run function func with input arguments self, the JIT_arrays
        if not len(self.other_arguments)==0:
            _ = self.func(self, *self.other_arguments, *self.JIT_Arrays)
        else:
            _ = self.func(self,*self.JIT_Arrays)
        if self.autosave:
            self.save()

    def calculate_all(self):
        any_done = False
        # find where the JIT_Arrays are not done
        is_not_done = ~self.JIT_Arrays[0].calculated
        # find indexes where the JIT_Arrays are not done
        ind = np.where(is_not_done)
        # construct array from indexes, by concatenation
        indexes = np.array(ind).T
        #indexes = np.empty((len(ind[0]), len(ind)), dtype=int)
        #for i, index in enumerate(ind):
        #indexes[:, i] = index
        self.calculate(indexes)

    def construct_vals_and_all_indexes(self, index):
        # first construct self.indexes_all, then vals
        ind = np.zeros(self.ndim_all, dtype=int)
        vals = []
        ind[self.array_vars_indexes] = index
        for i in range(self.ndim_all):
            vals.append(self.array_vars_all[i][ind[i]])
        self.indexes_all = ind
        self.curr_vals = vals

    def setup_iterator(self, indexes):
        self.indexes = indexes
        self.total_length = len(indexes)
        last_indexes = tuple(-np.ones(self.ndim, dtype=int))
        last_indexes_all = tuple(-np.ones(self.ndim_all, dtype=int))
        changed_var = np.ones(self.ndim_all, dtype=bool)
        if self.return_as_dict:
            self.last_indexes = last_indexes
            self.last_indexes_all = dict(zip(self.array_vars_all_names, last_indexes_all))
            self.changed_var = dict(zip(self.array_vars_all_names, changed_var))
        else:   
            self.last_indexes = last_indexes
            self.last_indexes_all = last_indexes_all
            self.changed_var = changed_var
        self.curr_index = -1

    def __next__(self):
        self.curr_index += 1
        curr_index = tuple(self.indexes[self.curr_index])
        # construct current directions
        self.construct_vals_and_all_indexes(curr_index)
        last_values = self.curr_vals
        changed_var = tuple(np.not_equal(self.indexes_all, self.last_indexes_all))
        if self.return_as_dict:
            self.last_values = dict(zip(self.array_vars_all_names, last_values))
            self.last_indexes = curr_index #dict(zip(self.array_vars_all_names, curr_index))
            self.last_indexes_all = self.indexes_all
            self.changed_var = dict(zip(self.array_vars_all_names, changed_var))
        else:   
            self.last_values = last_values
            self.last_indexes = curr_index
            self.last_indexes_all = self.indexes_all
            self.changed_var = changed_var
        return self.last_values, self.last_indexes, self.changed_var

    def kronprod(self, **args):
        self.set(**args)
        if self.do_tqdm and self.total_length > 1:
            self.pbar.init(self.indexes)
        if self.do_index:
            if self.do_change:
                for n in range(self.total_length):
                    v,i,c = next(self)
                    yield tuple(i), v, c
                    if self.do_tqdm and self.total_length > 1:
                        self.pbar.increment()
            else:
                for n in range(self.total_length):
                    v,i,_ = next(self)
                    yield tuple(i), v
                    if self.do_tqdm and self.total_length > 1:
                        self.pbar.increment()
        else:
            if self.do_change: 
                for n in range(self.total_length):
                    v,_,c = next(self)
                    yield v, c
                    if self.do_tqdm and self.total_length > 1:
                        self.pbar.increment()
            else:
                for n in range(self.total_length):
                    v,_,_ = next(self)
                    yield v
                    if self.do_tqdm and self.total_length > 1:
                        self.pbar.increment()
        if self.do_tqdm and self.total_length > 1:
            self.pbar.close()
            
    def tqdm(self, function, *iterator, weights=None, name='', **kwargs):
        if self.do_tqdm:
            return self.pbar.sub_tqdm(iterator, weights=weights, name=name, **kwargs)
        else:
            return weighted_tqdm(iterator, weights=weights, name=name, **kwargs)
        
    def p_tqdm(self, function, *iterator, weights=None, name='', **kwargs):
        if self.do_tqdm:
            return self.pbar.p_tqdm(function, *iterator, weights=weights, name=name, **kwargs)
        else:
            return weighted_p_tqdm(function, *iterator, weights=weights, name=name, **kwargs)

    def changed(self, elem=None):
        if elem is None:
            return self.changed_var
        elif isinstance(elem, int):
            if self.return_as_dict:
                string = self.array_vars_all_names[elem]
                return self.changed_var[string]
            else:
                return self.changed_var[elem]
        elif isinstance(elem, str): # Outputs changed by key
            if isinstance(self.array_vars_all_names, list):
                return self.changed_var[elem]
            else:
                raise ValueError('Keys are not defined, must create Object via dictionary for this functionality.')
                
    def index(self, elem=None):
        if elem is None:
            return self.last_indexes
        elif isinstance(elem, int):
            return self.last_indexes[elem]
        elif isinstance(elem, str): # By key
            if isinstance(self.array_vars_all_names, list):
                ind = self.array_vars_all_names.index(elem)
                return self.last_indexes_all[ind]
            else:
                raise ValueError('Keys are not defined, must create Object via dictionary for this functionality.')

    def value(self, elem=None):
        if elem is None:
            return self.last_values
        elif isinstance(elem, int):
            # is dictionary?
            if self.return_as_dict:
                string = self.array_vars_all_names[elem]
                return self.last_values[string]
            else:
                return self.last_values[elem]
        elif isinstance(elem, str):
            if isinstance(self.array_vars_all_names, list):
                return self.last_values[elem]
            else:
                raise ValueError('Keys are not defined, must create Object via dictionary for this functionality.')
    
    def output_definition_kronprod(self, **args):
        self.set(**args)
        i = 'index'
        v = 'value'
        c = 'change'
        if self.do_index:
            if self.do_change:
                return i, v, c
            else:
                return i, v
        else:
            if self.do_change: 
                return v, c
            else:
                return v, 

    # IF rng in func, then use Kron_Fun_Modifier to modify func and save it
    def func_modifier(self):
        func = self.func
        func_str = inspect.signature(func).parameters
        if 'rng' in func_str:
            data_dir = self.data_dir
            import_statements = self.import_statements
            fun_modifier = Kron_Fun_Modifier(func, self, data_dir, import_statements, other_func=self.other_func)
            func = fun_modifier.import_functions_from_file()
        self.func = func

    def all_combinations_array(self):
        # Generate an array for every combination of the input arrays
        # initialize arrays
        array_vars_all = []
        for i in range(self.ndim_all):
            curr_param = self.array_vars_all[i] # if curr_param doesn't have length, then make it an array
            if not hasattr(curr_param, '__len__'):
                curr_param = np.array([curr_param])
            curr_arr = np.empty(self.array_lengths_all, dtype=curr_param.dtype)
            # loop over all combinations
            for j in range(len(curr_param)):
                curr_arr[(slice(None), ) * i + (j,) ] = curr_param[j]
            array_vars_all.append(curr_arr)
        return array_vars_all



# -*- coding: utf-8 -*-
import numpy as np
import itertools
import warnings
import os


class Kron_Array(np.ndarray):   # Like JIT_Array but without saving and loading
    def __new__(cls, parent_shape, type, *args, **kwargs):
        # first args is an addition to shape, so we need to add it to the shape of the parent 
        if len(args) > 0: #parent.shape is a tuple
            if isinstance(args[0], tuple):
                shape_added = list(args[0])
            elif isinstance(args[0], int):
                shape_added = [args[0]]
            elif isinstance(args[0], list):
                shape_added = args[0]
            else:
                shape_added = []
            shape_parent = parent_shape
            shape = tuple(list(shape_parent) +shape_added)
            if len(args) > 1:
                args = args[1:]
            else:
                args = ()
            added_shape = len(shape_added)
        else:
            shape = parent_shape
            shape_parent = parent_shape
            added_shape = 0
        if type=='ones':
            obj = np.ones(shape, *args, **kwargs).view(cls)
        elif type=='zeros':
            obj = np.zeros(shape, *args, **kwargs).view(cls)
        elif type=='full':
            obj = np.full(shape, *args, **kwargs).view(cls)
        elif type=='empty':
            obj = np.empty(shape, *args, **kwargs).view(cls)
        else:
            raise ValueError('type not recognized')
        obj.added_shape = added_shape # did we add to the shape of the parent
        obj.shape_parent = shape_parent
        return obj
    
    def arg_index_formatting(self, arg):
        if self.added_shape > 0: # check if tuple with first element tuple
            if isinstance(arg, tuple) and isinstance(arg[0], tuple):
                arg = arg[0] + arg[1:]
            if isinstance(arg, tuple) and isinstance(arg[0], np.ndarray): # add arg1 to arg0 copy elements of arg1 to arg0
                # transform ndarray of arg[0] into tuple and append arg[1]
                arg = tuple(arg[0].flatten()) + arg[1:]
        if not isinstance(arg, tuple):
            if isinstance(arg, int):
                arg = (arg,)
            elif isinstance(arg, list):
                arg = np.array(arg)
            elif arg is Ellipsis:
                arg = (arg,)
            elif isinstance(arg, slice):
                arg = (arg,)
            else:
               arg = tuple(arg)
        return arg

    def __getitem__(self, arg):
        # add zero to *args to make sure that the first index is not a tuple
        arg = self.arg_index_formatting(arg)
        return super().__getitem__(arg).view(np.ndarray)
    
    def __setitem__(self, arg, value):
        # add zero to *args to make sure that the first index is not a tuple
        arg = self.arg_index_formatting(arg)
        # set the values
        super().__setitem__(arg, value)

    def asarray(self):
        return np.asarray(self)
    def array(self):
        return self.asarray()



# -*- coding: utf-8 -*-
import numpy as np
from weighted_tqdm import weighted_kronbinations_tqdm, weighted_tqdm

from hashlib import sha1
import os
import numpy as np


# An array class, that stores an array, and whether it's values have been calculated,
# and if so, what the values are
# If values of the array are requested, and they have not been calculated, they are calculated using a function handle passed to the constructor
class JIT_kronbinations_load():
    def __init__(self, *values, func=None, import_statements=[], other_arguments=[], checksum=None, checksum_pre='', data_dir='Cache', **kwargs):
        # Calculate checksums
        self.data_dir = data_dir
        # check if data_dir exists
        if not os.path.exists(data_dir):
            os.makedirs(data_dir)
        if checksum is None:
            self.checksum = self.make_checksum(checksum_pre, func, *values, *import_statements, *other_arguments)
        else:
            self.checksum = checksum
        # Check if files with checksum exist
        filenames, var_names = self.filenames()
        n_filenames = len(filenames)
        if n_filenames == 0:
            raise ValueError('No files with checksum ' + self.checksum + ' found!')
        # Load files
        self.vars = {}
        self.var_calculated = {}
        for i, (filename, var_name) in enumerate(zip(filenames, var_names)):
            new_var, new_calculated, _ = self.load_filename(filename)
            self.vars[var_name] = new_var
            self.var_calculated[var_name] = new_calculated

        # lengths of the values -> ignore the length one arrays, they are not to be iterated over
        # if values is a dictionary, then the keys are the directions, and the values are the arrays
        if isinstance(values[0], dict):
            if len(values) > 1:
                raise ValueError('If values is a dictionary, it must be the only argument')
            input_values = values[0]
            keys = input_values.keys()
            values = input_values.values()
            self.array_vars_all_names, self.array_vars_all, self.all_weights = [], [], []
            for i, (key, value) in enumerate(zip(keys, values)):  
                # check if value is a dict with value and weight
                self.array_vars_all_names.append(key)
                if isinstance(value, dict):
                    if 'value' in value.keys():
                        self.array_vars_all.append(value['value'])
                    else:
                        raise ValueError('If value is a dictionary, it must have a value key') 
                else: # Normal 
                    self.array_vars_all.append(value)
            self.return_as_dict = True
        else:
            self.array_vars_all = list(values)
            self.return_as_dict = False
            self.array_vars_all_names = None
        for i, arr in enumerate(self.array_vars_all):
            # if does not have length, transform into array
            if not hasattr(arr, '__len__'):
                self.array_vars_all[i] = np.array([arr])
        # only relevant values in array_directions
        self.array_vars = [arr for arr in self.array_vars_all if len(arr) > 1]
        if isinstance(values, dict):
            self.array_vars_names = [name for name, arr in zip(self.array_vars_all_names, self.array_vars_all) if len(arr) > 1]
        # add index values of the array vars 
        self.array_vars_indexes = [i for i, arr in enumerate(self.array_vars_all) if len(arr) > 1] 

        if self.return_as_dict:
            self.curr_vals = {key: arr[0] for key, arr in zip(self.array_vars_all_names, self.array_vars_all)}
        else:
            self.curr_vals = [arr[0] for arr in self.array_vars_all]
        self.array_lengths_all = [len(arr) for arr in self.array_vars_all]
        self.array_lengths = [len(arr) for arr in self.array_vars]
        self.shape_all = tuple(self.array_lengths_all)
        self.shape = tuple(self.array_lengths)
        self.ndim_all = len(self.array_lengths_all)
        self.ndim = len(self.array_lengths)
        self.size_all = np.prod(self.array_lengths_all)
        self.size = np.prod(self.array_lengths)
        self.set(**kwargs)   # redo these values if passed as kwargs
        
    def get_arrays(self):
        return self.vars
    
    def get_calculated(self):
        return self.var_calculated
    
    def set(self, **args):
        key_substitution_list = [['index', 'do_index'], ['change', 'do_change'], ['progress', 'do_tqdm']]
        key_list = [v[0] for v in key_substitution_list]
        subs_list = [v[1] for v in key_substitution_list]
        for key, value in args.items():
            # Substitute certain keys from substitution list
            if key in key_list:
                key = subs_list[key_list.index(key)]
            if (key == 'return_as_dict' and value==True) and not isinstance(self.array_vars_all_names, list):
                raise ValueError('Keys are not defined, must create Object via dictionary in order to set "return_as_dict = True".')
            else:
                setattr(self, key, value)

    def get(self, *args):
        key_substitution_list = [['index', 'do_index'], ['change', 'do_change'], ['progress', 'do_tqdm']]
        key_list = [v[0] for v in key_substitution_list]
        subs_list = [v[1] for v in key_substitution_list]
        x = []
        for key in args:
            if key in key_list:
                key = subs_list[key_list.index(key)]
            x.append(getattr(self, key))
        if len(x) == 1:
            return x[0]
        else:
            return x
        
    def __getitem__(self, key):
        # If the key is not in the data, return None
        if key in key_list:
            key = subs_list[key_list.index(key)]
        if key not in self.data:
            return None
        else:
            return self.data[key]

    def load_filename(self, filename):
        try:
            with np.load(os.path.join(self.data_dir, filename), allow_pickle=True) as f:
                var = f['var']
                calculated = f['calculated']
                how_many_missing = f['how_many_missing']
        except:
            raise Exception('Error loading file: ' + filename)
        return var, calculated, how_many_missing
            
    def filenames(self):
        # searches for filenames in data_dir with checksum and ending on .npz
        filenames = []
        var_names = []
        for file in os.listdir(self.data_dir):
            if file.endswith('.npz') and file.startswith(self.checksum):
                filenames.append(file)
                new_file = file.replace(self.checksum + '_', '')
                new_file = new_file.replace('.npz', '')
                var_names.append(new_file)
        return filenames, var_names
    
    def change_all_dtype(self, object, dtype='int'):
        # navigates through the objects lists tuples and dictionaries until it finds a numpy array 
        # and changes the dtype if it is an integer to 'int'
        if isinstance(object, (list, tuple, dict)):
            # recursively call the function on the elements of the object
            is_tuple = isinstance(object, tuple)
            if is_tuple:
                object = list(object)
            if isinstance(object, dict):
                # navigate every key value pair
                for key, value in object.items():
                    object[key] = self.change_all_dtype(value, dtype=dtype)
            else: # isinstance(object, list):
                for i in range(len(object)):
                    object[i] = self.change_all_dtype(object[i], dtype=dtype)
            if is_tuple:
                object = tuple(object)
        elif isinstance(object, np.ndarray):
            if 'int' in str(object.dtype):
                object = object.astype(dtype)
        return object
    
    def make_checksum(self, checksum_pre, func, *args):
        # check every arg in args, if arg is an object of a class 
        # change all dtype int
        args = self.change_all_dtype(args)
        checksum_sha1 = sha1(str(args).encode('utf-8')).hexdigest()
        checksum_sin_fun = checksum_pre + checksum_sha1
        # get function name from func
        if func is not None:
            func_name = func.__name__
            checksum = checksum_pre + checksum_sha1 + func_name
            # check if files with checksum exist
            checksum_exists = False
            dir_list = os.listdir(self.data_dir)
            for file in dir_list:
                if checksum in file:
                    checksum_exists = True
                    break
            if not checksum_exists:
                # check if files with checksum_sin_fun exists -> rename to checksum
                for file in dir_list:
                    if checksum_sin_fun in file:
                        new_filename = file.replace(checksum_sin_fun, checksum)
                        os.rename(os.path.join(self.data_dir, file), os.path.join(self.data_dir, new_filename))  
        else:
            checksum = checksum_sin_fun
        return checksum



import pytest
import numpy as np
from kronbinations import kronbinations

# Generate lists of different dtypes
# and test the functions on them
floats = np.array([1.2, 3.4, 5.6, 7.8, 9.0])
ints = np.array([1, 3, 5, 7, 9], dtype=int)
strings = np.array(['a', 'b', 'c', 'd', 'e'])

# Create a kronbinations object
def test_kronbinations_matrices():
    # Create a kronbinations object
    # with the lists as values
    k = kronbinations(floats, ints, strings)
    assert k.size == 125
    assert k.shape == (5,5,5)
    assert k.ndim == 3
    # Test the generation of matrices
    o = k.ones(dtype=int)
    z = k.zeros(dtype=int)
    e = k.empty()
    f = k.full(fill_value= 1.0/3.0)
    assert o.shape == (5,5,5)
    assert z.shape == (5,5,5)
    assert e.shape == (5,5,5)
    assert f.shape == (5,5,5)
    
def test_kronbinations_loop_outputs():
    k = kronbinations(floats, ints, strings)
    b = [False, True]
    # Check number of outputs
    for index in b:
        for change in b:
            for progress in b:
                how_many = 1 if index else 0
                how_many += 1 if change else 0
                if how_many == 0:
                    how_many = 2 # elements of the values tuple - 1
                for output in k.kronprod(index=index, change=change, progress=progress):
                    assert len(output) == how_many+1

def test_krombinations_intermediate_outputs():
    # Check the intermediate outputs
    k = kronbinations(floats, ints, strings)
    for i,v,c in k.kronprod(index=True, change=True, progress=True):
        assert k.changed() == c
        assert k.index() == i
        assert k.value() == v
        for j in range(3):
            assert k.changed(j) == c[j]
            assert k.index(j) == i[j]
            assert k.value(j) == v[j]

def test_krombinations_settings_and_changes_to_them():
    k = kronbinations(floats, ints, strings)
    k.set(do_index=True, do_change=True, do_tqdm=True)
    do_index, do_change, do_tqdm, return_as_dict = k.get('do_index', 'do_change', 'do_tqdm', 'return_as_dict')
    assert do_index
    assert do_change
    assert do_tqdm
    k.set(do_index=False, do_change=False, do_tqdm=False)
    do_index, do_change, do_tqdm, return_as_dict = k.get('index', 'do_change', 'do_tqdm', 'return_as_dict')
    assert not do_index
    assert not do_change
    assert not do_tqdm

def test_kronbinations_illegal_dict_outputs():
    # Check the illegal dictionary outputs
    k = kronbinations(floats, ints, strings)
    with pytest.raises(ValueError):
        for i,v,c in k.kronprod(index=True, change=True, progress=False):
            k.value('floats')
    with pytest.raises(ValueError):
        for i,v,c in k.kronprod(index=True, change=True, progress=False):
            k.index('floats')
    with pytest.raises(ValueError):
        for i,v,c in k.kronprod(index=True, change=True, progress=False):
            k.changed('floats')

def test_kronbinations_dict_objects():    
    # Create a kronbinations object
    # with the lists as a dictionary
    keys = ['floats', 'ints', 'strings']
    d = {'floats': floats, 'ints': ints, 'strings': strings}
    k = kronbinations(d)
    with pytest.raises(ValueError):
        k = kronbinations(d, d)
    for i,v,c in k.kronprod(index=True, change=True, progress=False):
        for j, key in enumerate(keys):
            assert k.changed(key) == c[key]
            assert k.index(key) == i[j]
            assert k.value(key) == v[key]
            


import inspect
import importlib
import re # regular expressions
import os
import numpy as np

class Kron_Fun_Modifier:
    # Modifies function handles containing kronbinations and kronprod terms, modifies them for use with JIT_Arrays
    def __init__(self, func, parent, data_dir='Cache', import_statements=None, other_func=[]):
        self.func = func
        self.other_func = other_func
        self.import_statements = import_statements
        self.parent = parent
        self.data_dir = data_dir

        self.func_str = inspect.getsource(self.func)
        # separate semicolon separated statements into separated lines, and unify multi line statements inside of brackets
        self.func_str = self.simple_reformatting()
        # Separate function into def, body, name, input arguments and output arguments
        self.definition, self.body, self.name, self.input_args = self.separate()  # , self.return_args, self.number_of_return_statements
        self.indentation_depths, self.begin_indentation, self.end_indentation, self.indentation_types = self.classify_indentations()
        self.all_variables, self.first_appearance = self.find_all_variables()
        # Find variables by line
        # go line by line and find variables by line
        self.body_classification = {'variables': [], 'where': [], 'indexed': [], 'defined': []}
        for i, line in enumerate(self.body):
            variables, where, indexed, defined = self.variables_in_line(line)
            self.body_classification['variables'].append(variables)
            self.body_classification['where'].append(where)
            self.body_classification['indexed'].append(indexed)
            self.body_classification['defined'].append(defined)
        # separate body into intro and loop
        _, _, self.kronbinations_index = self.before_after()
        #do_rng = self.contains_rng()
        string = self.body[self.kronbinations_index]
        for_variables, for_variable_types = self.classify_for_variables_kronprod(string)
        self.new_body = self.rng_conditions_stack(for_variables, for_variable_types)
        #print(self.construct_function_string())
        self.file_name = self.save()

    def save(self):
        file_name = os.path.join(self.data_dir, self.name + '.py')
        with open(file_name, 'w') as f:
            f.write(self.construct_function_string())
        # return file_name
        return file_name

    def before_after(self, input_lines=None, separator='kronprod', *lists):
        # separates function into intro with variable definitions and loop with kronbination
        # detect kronprod as separator
        # detect kronbination as initiator
        if input_lines is None:
            input_lines = self.body
        n_lines = len(input_lines)
        line_index = -1
        curr_index = 0
        line_gen = iter(input_lines)
        while line_index < 0 and curr_index < n_lines:
            curr_line = next(line_gen)
            if separator in curr_line:
                line_index = curr_index
            curr_index += 1
        if line_index < 0:
            raise ValueError('Separator not found')
        before = input_lines[:line_index]
        after = input_lines[line_index:]
        # separate *lists into before and after
        if len(lists) > 0:
            before_lists = []
            after_lists = []
            for l in lists:
                before_lists.append(l[:line_index])
                after_lists.append(l[line_index:])
            return before, after, before_lists, after_lists
        return before, after, line_index
        

    def simple_reformatting(self):
        # reformat statements so that ; separated statements are on separate lines and multi line statements are on one line
        func_str = self.func_str
        # separate string into a list of strings by line   
        func_str_lines = func_str.splitlines()
        # remove empty lines
        func_str_lines = [line for line in func_str_lines if line.strip() != '']
        # remove comments
        new_func_str_lines = []
        for line in func_str_lines:
            if '#' in line:
                remaining_str = line.split('#')[0]
                if remaining_str.strip() != '':
                    new_func_str_lines.append(remaining_str)
            else:
                new_func_str_lines.append(line)
        func_str_lines = new_func_str_lines
        # remove docstrings
        new_func_str_lines = []
        currently_in_docstring = False
        for line in func_str_lines:
            if "'''" in line:
                # check if appears twice in the line
                if not line.count("'''") == 2:
                    currently_in_docstring = not currently_in_docstring   # Toggle
                    # keep start of the string
                    if currently_in_docstring:
                        # docstring is starting
                        #keep beginning of the string
                        beginning = line.split("'''")[0]
                        if beginning.strip() != '':
                            new_func_str_lines.append(beginning)
            elif not currently_in_docstring:
                new_func_str_lines.append(line)
        func_str_lines = new_func_str_lines
        # separate semicolon separated statements into separate lines
        new_func_str_lines = []
        for line in func_str_lines:
            if ';' in line:
                new_func_str_lines += line.split(';')
            else:
                new_func_str_lines.append(line)
        func_str_lines = new_func_str_lines

        # unify multi line statements inside of brackets, find open brackets that aren't closed
        new_func_str_lines = []
        currently_in_brackets = np.array([0,0,0])
        open_braces = ['(', '[', '{']
        close_braces = [')', ']', '}']
        last_in_brackets = False
        for line in func_str_lines:
            if last_in_brackets == False:
                curr_line = line.rstrip() 
            for i, brace in enumerate(open_braces):
                currently_in_brackets[i] += line.count(brace)
            for i, brace in enumerate(close_braces):
                currently_in_brackets[i] -= line.count(brace)
            if np.any(currently_in_brackets < 0):
                raise ValueError('Too many closing brackets')
            
            if np.sum(currently_in_brackets):
                if last_in_brackets == True:
                    curr_line += ' ' + line.strip()
                    last_in_brackets = True
            else:
                if last_in_brackets == True:
                    curr_line += ' ' + line.strip()
                new_func_str_lines.append(curr_line)
                last_in_brackets = False
            if np.any(currently_in_brackets > 0):
                last_in_brackets = True
        func_str_lines = new_func_str_lines
        # combine lines into one string
        func_str = ''
        for line in func_str_lines:
            func_str += line + '\n'
        return func_str
    
    # Separate function into def, body, name, input arguments and output arguments
    def separate(self):
        func_str = self.func_str
        # separate string into a list of strings by line
        func_str_lines = func_str.splitlines()
        definition = func_str_lines[0]
        body = func_str_lines[1:]
        ### name is the first word after def
        name = definition.split()[1].split('(')[0]
        ### input arguments
        input_args_unformatted = definition.split('(')[1].split(')')[0]
        # remove equal signs, the default values after equal signs and remove spaces
        input_args = [arg.split('=')[0].strip() for arg in input_args_unformatted.split(',')]
        return definition, body, name, input_args

    # get indentation depth of each line in the body, beginning of indentations, ends and types
    def indentation_blocks(self, indentations_by_line):
        """Return a list of tuples of the form (start_line, end_line) for each block of code."""
        # indentations_by_line is an array of the indentation level of each line (int)
        # ind start is an index to be added to the start of the block and the end of the block
        # do it iteratively
        start = []
        ender = []
        len_ind = len(indentations_by_line)
        curr_ind = -1
        for i in range(1, len_ind):
            if indentations_by_line[i] > indentations_by_line[i-1]:
                curr_indentations = indentations_by_line[i]
                start.append(i - 1)
                # find end of block
                for j in range(i+1, len_ind):
                    if indentations_by_line[j] < curr_indentations:
                        ender.append(j)
                        break
        return start, ender
    def classify_indentations(self, input_lines=None):
        if input_lines is None:
            input_lines = self.body
        indentation_depths = np.array([len(line) - len(line.lstrip()) for line in input_lines])
        lines = [line.strip() for line in self.body]
        tab_to_spaces = np.min(indentation_depths)
        #print(indentation_depths)
        indentation_depths = indentation_depths // tab_to_spaces
        begin_indentation, end_indentation = self.indentation_blocks(indentation_depths)
        indentation_types = []
        for i in range(len(begin_indentation)):
            curr_type = lines[begin_indentation[i]].strip().split()[0]
            if curr_type == 'else:':
                curr_type = 'else'
            indentation_types.append(curr_type)
        return indentation_depths, begin_indentation, end_indentation, indentation_types

        # function to detect an equal sign in a line
    def brace_depth(self, line):
        open_braces = ['(', '[', '{']
        closed_braces = [')', ']', '}']
        all_braces = open_braces + closed_braces
        # for every char get the brace level and store whether it is a brace
        n = len(line)
        brace_levels = - np.zeros(n, dtype=int)
        brace_types = - np.ones(n, dtype=int)
        last_brace = []
        curr_brace_level = 0
        for i, char in enumerate(line):
            # New brace? 
            if curr_brace_level < 0:
                print('Error: brace level is negative')
            if char in all_braces:
                if char in open_braces:
                    curr_brace_level += 1
                    last_brace.append(open_braces.index(char))
                    # don't store brace level and type during a brace
                elif char in closed_braces:
                    curr_brace_level -= 1
                    # delete last brace element
                    last_brace.pop()
                    # don't store brace level and type during a brace
                brace_levels[i] = -1
            else:
                brace_levels[i] = curr_brace_level
                if curr_brace_level > 0:
                    brace_types[i] = last_brace[-1]      
        return brace_levels, brace_types
    def detect_equal_sign(self, line):
        # ignore equal signs in braces
        if not '=' in line:
            return 0
        else:
            # index of first occurance
            index = line.find('=')
            prev_char = line[index-1]
            # occurance must also not be !=, ==, >=, <=, 
            list_of_equal_signs = ['!', '>', '<']
            if prev_char in list_of_equal_signs:
                return 0
            next_char = line[index+1]
            if next_char == '=':
                return 0
            # occurance must be at brace_level = 0
            brace_levels, brace_types = self.brace_depth(line[:index])
            if brace_levels[-1] > 0:
                return 0
            return index
    def detect_for_statement(self, line):
        # ignore equal signs in braces
        if not 'for ' in line:
            return False, [], []
        else:
            # index of first occurance
            index_for = line.find('for ')
            index_in = line[index_for:].find(' in ')+index_for
            # is there an in statement after the for statement
            if not 'in' in line[index_for:]:
                print("A ' for ' statement must be followed by an ' in ' statement: ", line)
                return False, [], []
            else:
                # find first index after for statement and left index before in statement for every for ... in ... statement in line
                left =  [index_for+3]
                right = [index_in]
                # find other occurances
                while 'for' in line[index_for+1:]:
                    index_for = line[index_for+1:].find('for ')+index_for+1
                    index_in = line[index_in+1:].find(' in ')+index_in+1
                    left.append(index_for+3)
                    right.append(index_in)
            return True, left, right
    # function to separate comma separated variables
    def separate_variables(self, variable):
        # separate by commas or braces
        # first remove braces ()
        if '(' in variable:
            # check if ) in variable
            if ')' in variable:
                # are there variables before   -> does not yet support partial tuple unpacking, eg. a, (b, c) = d
                # replace ( by ,
                variable = variable.strip()
                while variable[0] == '(':
                    variable = variable[1:]
                while variable[-1] == ')':
                    variable = variable[:-1]
                # remove the other braces
                variable = variable.replace('(','')
                variable = variable.replace(')', '')
                variable = variable.replace(' ','')
            else:
                don_t_add = True
                print('Warning: could not parse line: '+variable, ' (missing closing parenthesis)')
        # last thing, separate multiple variable assignment by commas
        if ',' in variable:
            variable = variable.split(',')
            variable = [var.strip() for var in variable]
        if isinstance(variable, str):
            variable = [variable]
        return variable
    # find all variables in function
    def find_all_variables(self):
        # find all variables in function (variables) and their first appearances (first_appearances)
        # get variables in input arguments
        variables = [arg for arg in self.input_args]
        first_appearances = [-1 for i in range(len(self.input_args))]
        # function to remove indexes from variables a[1] --> a
        def remove_indexes(variable):
            if '[' in variable:
                variable = variable.split('[')[0].strip()
            return variable
        def extract_variables(string):
            # , separated and remove ()
            # remove indexes
            vars = string.replace('(','').replace(')','').strip().split(',')
            return [remove_indexes(var.strip()) for var in vars]
        for i, l in enumerate(self.body):
            line = l.strip()
            # find equal signs
            equal_sign = self.detect_equal_sign(line)
            # find for statements
            for_statement, left, right = self.detect_for_statement(line)
            # find all variables in these statements
            if equal_sign:
                # get variables before
                curr_variables = extract_variables(line[:equal_sign])
                curr_appearances = [i for j in range(len(curr_variables))]
            elif for_statement:# only consider first for statement
                curr_variables = extract_variables(line[left[0]:right[0]])
                curr_appearances = [i for j in range(len(curr_variables))]
            else:
                curr_variables = []
                curr_appearances = []
            # add variables to list
            variables += curr_variables
            first_appearances += curr_appearances

        # make variables unique, but remove same elements in first_appearance
        #for var, first in zip(variables, first_appearances):
        #    print(var, first)
        variables, indexes = np.unique(variables, return_index=True)
        first_appearances = [first_appearances[i] for i in indexes]
        # sort appearances anmd get sorting indexes
        first_appearances, indexes = np.sort(first_appearances), np.argsort(first_appearances)
        variables = [variables[i] for i in indexes]
        return variables, first_appearances


    def variables_in_line(self, line):
        if isinstance(line, int):
            line = self.body[line]
        variable_list = self.all_variables
        # find all variables in line
        def remove_equal_sign(line):
            # a string sthat should be separated at an equal sign
            # relevant part is after equal sign remove everything before it
            return line[line.find('=')+1:].strip()
        def is_indexed(string, end_index):
            # checks if a variable is indexed and removes the index, check for  [
            n = len(string)
            curr_index = end_index    
            while curr_index < n:
                if string[curr_index] == ' ': # ignore spaces
                    curr_index += 1
                elif string[curr_index] == '[': # found index
                    return True
                else:
                    return False
        line = line.strip()
        # find equal signs
        equal_sign = self.detect_equal_sign(line)
        # find for statements
        for_statement, left, right = self.detect_for_statement(line)
        # brace levels and types
        brace_levels, brace_types = self.brace_depth(line)
        variables = []
        indexed = []
        defined = []    
        where = []
        # find all variables in these statements
        separators = ['=', '+', '-', '*', '/', '>', '<', ',', '.', '%', '&', '|', '^', '~', '!', '?', ':', ';', '(', ')', '[', ']', '{', '}', ' '] #symbols that can end or start a variable name
        for var in variable_list:
            # are there any occurances of var in line
            if var in line:
                # find all occurances of var in line
                indexes = [m.start() for m in re.finditer(var, line)]
                n = len(var)
                # check if var is indexed var[...
                curr_vars = []
                curr_where = []
                curr_indexed = []
                curr_defined = []
                for i in indexes:
                    # is this a true hit or just a substring of a longer function or variable name?   -> compare adjacent characters to separators
                    if i > 0: # to the left
                        if line[i-1] not in separators:
                            continue # skip this iteration
                    if i+n < len(line): # to the right
                        if line[i+n] not in separators:
                            continue
                    # if we get here, we have a true hit, still need to classify this variable
                    curr_vars.append(var)
                    curr_where.append(i)
                    curr_is_indexed = is_indexed(line, i+n)
                    curr_indexed.append(curr_is_indexed)
                    # is defined                
                    if curr_is_indexed:
                        curr_defined.append(False)
                    else:
                        # left of equal sign?
                        if equal_sign > i:
                            if brace_types[i] == 1:  # is within square brackets []
                                curr_defined.append(False)
                            else:
                                curr_defined.append(True)
                        elif for_statement: # is within for...in statement?
                            cdef = False
                            j = 0
                            while j <len(left) and cdef == False:
                                if left[j] < i and i < right[j]:
                                    cdef = True
                                j += 1
                            curr_defined.append(cdef)
                        else:
                            curr_defined.append(False)
                variables += curr_vars
                where += curr_where
                indexed += curr_indexed
                defined += curr_defined
        # sort by location in line (where) from right to left 
        indexes =np.argsort(where)[::-1]
        where = [where[i] for i in indexes]
        variables = [variables[i] for i in indexes]
        indexed = [indexed[i] for i in indexes]
        defined = [defined[i] for i in indexes]

        return variables, where, indexed, defined   
    
    def remove_from_definition(self, var):
        # remove the variable var from the definition (first line of function)
        definition = self.definition
        # find variables in all_variables, match first_appearance
        if var in self.all_variables:
            ind = self.all_variables.index(var)
            first = self.first_appearances[ind]
            if first >= 0: # is defined in function
                print('Variable not found in function definition')
                return definition
            # remove variable from definition
            index_left = definition.find('(')
            # find after index_left
            index_var = definition.find(var, index_left) 
            len_var = len(var)
            # remove from string
            new_definition = definition[:index_var] + definition[index_var+len_var:]
            return new_definition
        else:
            print('Variable not found in function definition')
            return definition

    # use rng?
    def contains_rng(self, line_indexes=None):  #line_indexes = [min, max] or simply index
        all_variables = self.all_variables
        first_appearances = self.first_appearance
        if not 'rng' in all_variables:
            return False
        else:
            if line_indexes is None:
                if first_appearances[all_variables.index('rng')] < 0:
                    return True
                else:
                    return False
            else:
                if isinstance(line_indexes, int):
                    variables, _, _, _ = self.variables_in_line(line_indexes)
                    if 'rng' in variables:
                        return True
                else:
                    for i in range(line_indexes[0], line_indexes[1]):
                        # variables in line
                        variables, _, _, _ = self.variables_in_line(i)
                        if 'rng' in variables:
                            return True
                return False
    
    def classify_for_variables_kronprod(self, string):
        def list_of_arguments(string):
            string=string.replace(' ', '') # remove spaces
            # separate string iinto a list of strings, and elements within braces into subslists of strings, so that a, (b, c) becomes ['a', ['b', 'c']]
            # separate string into a list of different brace depths, so that a, (b, c) becomes ['a', ['b, c']]
            
            # find braces first
            brace_depth = 0
            brace_depths = np.empty(len(string), dtype=int)
            for i, c in enumerate(string):
                if c == '(':
                    brace_depth += 1
                elif c == ')':
                    brace_depth -= 1
                brace_depths[i] = brace_depth
            # now split the string into a list of strings, each with the same brace depth
            curr_depth = 0
            curr_string = ''
            strings = []
            for i, c in enumerate(string):
                if brace_depths[i] == curr_depth:
                    curr_string += c
                elif brace_depths[i] < curr_depth:
                    strings.append( [curr_string] )
                    curr_string = ''
                    curr_depth = brace_depths[i]
                else: #if brace_depths[i] > curr_depth:
                    strings.append(curr_string)
                    curr_string = ''
                    curr_depth = brace_depths[i]
            else:
                strings.append(curr_string)
            # now split by commas, strip and remove empty strings
            def clean_str_list(str_list):
                new_str_list = []
                for s in str_list:
                    if isinstance(s, str):
                        s = s.strip()
                        if s != '':
                            s = s.replace(',', '')	
                            s_split = s.split(' ')
                            s_split = [s.strip() for s in s_split]
                            new_str_list +=s
                    else:
                        new_str_list.append(clean_str_list(s))
                return new_str_list
            return clean_str_list(strings)
        _, left, right = self.detect_for_statement(string)
        sub_string = string[left[0]:right[0]].strip()
        variables = list_of_arguments(sub_string)
        # get arguments to detect changed indexing, change, progress behaviour
        argument_string = string[right[0]:].strip()
        #find first '(' and last ')'
        left = argument_string.find('(')
        right = argument_string.rfind(')')
        argument_string = argument_string[left+1:right]
        # split by comma
        arguments = argument_string.split(',')
        arguments = [a.strip().split('=') for a in arguments]
        def str_2_bool(s):
            return True if s == 'True' else False
        argument_dict = {key.strip(): str_2_bool(value.strip()) for key, value in arguments}
        names = self.parent.output_definition_kronprod(**argument_dict)
        if not isinstance(names, tuple):
            names = (names, )
        for_variables = []
        for_variable_types = []
        if not len(names) == len(variables):
            raise ValueError('kronprod returns {} variables, but {} variables are defined in for loop'.format(len(names), len(variables)))
        for i, name in enumerate(names):
            for_variables += variables[i]
            for_variable_types += [name for var in variables[i]]
        return for_variables, for_variable_types
    #def replace_braces_outside_of_function_calls(self, string):
    def change_kronbinations_calls(self, string):
        # change k.changed() to k.value()
        # first find variable name of kronbinations_object in kronprod line
        kronbinations_index = self.kronbinations_index
        line = self.body[kronbinations_index]
        # find variable name
        var_name = line.split('.kronprod')[0].strip().split(' ')[-1]
        # find var_name.changed() and replace with var_name.value() 
        # find all instances of var_name.changed()
        changed = var_name + '.changed('
        value = var_name + '.value('
        index = var_name + '.index('
        all_three = [changed, value, index]
        # find all instances:
        var_list = []
        for i in range(len(all_three)):
            curr_var = all_three[i]
            # find all occurances of curr_var in string one after another
            new_ind = string.find(curr_var)
            ind_begin = []
            while new_ind >= 0:
                ind_begin.append(new_ind)
                new_ind = string.find(curr_var, new_ind+len(curr_var))
            #ind_begin = [m.start() for m in re.finditer(curr_var, string)]
            # find end of each instance
            ind_end = []
            for j in range(len(ind_begin)):
                ind = ind_begin[j]
                #check if the character before ind_begin is a letter or underscore, if not, then this is a call
                if string[ind-1] != '_' and not string[ind-1].isalpha():
                    # find the next )
                    ind_end.append(string.find(')', ind))
                else:
                    #remove from ind_begin
                    ind_begin[j] = -1
            # remove -1's
            ind_begin = [ind_begin[i] for i in range(len(ind_begin)) if ind_begin[i] >= 0]
            # remove all hits from string and add them to a variable list
            for j in range(len(ind_begin)):
                begin = ind_begin[j]
                ender = ind_end[j]
                curr_string = string[begin:ender+1]
                string = string[:begin] + ' '*len(curr_string) + string[ender+1:]
                if i == 0:
                    curr_string = curr_string.replace('changed', 'value')
                var_list.append(curr_string)
        return string, var_list

    def rng_conditions_stack(self, for_variables, for_variable_types):
        # find all if statements in body
        # check if rng variable exists in body
        seed = 100
        indentation_depths = self.indentation_depths
        begin_indentation = self.begin_indentation 
        end_indentation = self.end_indentation
        indentation_types = self.indentation_types
        if_statements_indexes = []
        body = self.body.copy()
        len_body = len(body)
        # which if statements are rng'ed if statements
        for i in range(len(indentation_types)):
            if indentation_types[i] == 'if' or indentation_types[i] == 'elif' or indentation_types[i] == 'else':
                # check if there is a rng statement in the if statement
                begin = begin_indentation[i]
                ender = end_indentation[i]
                if self.contains_rng([begin, ender]):
                    if_statements_indexes.append(i)
        
        # analyze line by line which conditions are in the current indentation level
        if len(if_statements_indexes) > 0:
            begin = begin_indentation[if_statements_indexes[0]]
            conditions_by_line = [[] for i in range(len_body)]
            kronbinations_index = self.kronbinations_index
            for i in if_statements_indexes:
                begin = begin_indentation[i]
                ender = end_indentation[i]
                if begin > kronbinations_index:
                    condition_line = body[begin]
                    conditions_string = condition_line.strip()
                    if conditions_string.startswith('if'):
                        conditions_string = conditions_string[2:]
                    elif conditions_string.startswith('elif'):
                        conditions_string = conditions_string[4:]
                    elif conditions_string.startswith('else'):
                        conditions_string = conditions_string[4:]
                    if conditions_string.endswith(':'):
                        conditions_string = conditions_string[:-1]
                    conditions_string = ' '+conditions_string+' '
                    conditions_string, var_list = self.change_kronbinations_calls(conditions_string)
                    conditions_string = conditions_string.replace('(', ' ').replace(')', ' ')  
                    sep_list = [' and ', ' not ', ' or ', ' in ', '==', '!=', '<=', '>=', '<', '>', ' ']
                    for sep in sep_list:
                        conditions_string = conditions_string.replace(sep, ',')
                    # split at commas, keep non empty strings
                    conditions = [c.strip() for c in conditions_string.split(',') if c.strip() != '']  
                    # conditions without indexing
                    conditions_no_index = []
                    for c in conditions:
                        if '[' in c:
                            conditions_no_index.append(c[:c.find('[')])
                        else:
                            conditions_no_index.append(c)
                    # match with for variables # substitute change with var
                    for k, (c, c_no_ind) in enumerate(zip(conditions, conditions_no_index)):
                        if c_no_ind in for_variables:
                            # get index of variable in for_variables
                            index = for_variables.index(c_no_ind)
                            # get type of variable
                            var_type = for_variable_types[index]
                            # if var_type is change -> substitute with val variable
                            if var_type == 'change':
                                # how many_change before this index?
                                n_changes = for_variable_types[:index+1].count('change')
                                # get the val nth var variable where n is n_changes
                                val_var = None
                                curr_n_changes = n_changes
                                for j, v in enumerate(for_variables):
                                    if for_variable_types[j] == 'value':
                                        curr_n_changes -= 1
                                        if curr_n_changes == 0:
                                            val_var = v
                                            break
                                if val_var is None:
                                    # try exchanging with index variable
                                    ind_var = None
                                    curr_n_changes = n_changes
                                    for j, v in enumerate(for_variables):
                                        if for_variable_types[j] == 'index':
                                            curr_n_changes -= 1
                                            if curr_n_changes == 0:
                                                ind_var = v
                                                break
                                    if ind_var is None:
                                        raise Exception('value or index variables corresponding to change variable ('+c+') not found')
                                    else:
                                        conditions[k] = ind_var + c[c.find('['):]
                                        conditions_no_index[k] = ind_var
                                else:
                                    conditions[k] = val_var + c[c.find('['):]
                                    conditions_no_index[k] = val_var
                    # add current conditions to var_dict
                    # add conditions to conditions_by_line for elements begin to end
                    conditions += var_list
                    for k in range(begin+1, ender):
                        conditions_by_line[k] += conditions
            # check line by line for a rng statement
            rng_in_line = [False for i in range(len_body)]
            for i, line in enumerate(body):
                # find variables in line
                if self.contains_rng(i):
                    rng_in_line[i] = True
            # for every indentation check if there is a rng statement and, that isn't in a sub rng statement
            rng_in_if_statement = []
            for i in if_statements_indexes:
                begin = begin_indentation[i]
                ender = end_indentation[i]
                if any(rng_in_line[begin:ender]):
                    # where is the rng statement?
                    rng_index = [i+begin for i, x in enumerate(rng_in_line[begin:ender]) if x]
                    rng_in_if_statement.append(rng_index)
                else:
                    rng_in_if_statement.append([])
            # which indexes are unique to one if statement and not in a sub if statement
            all_rng_indexes = [i for i, x in enumerate(rng_in_line) if x]
            unique_rng_in_if_statement = [[] for i in range(len(if_statements_indexes))]
            for index in all_rng_indexes:
                n_if = len(if_statements_indexes)
                for rng_in_if in rng_in_if_statement[::-1]: # last occurance is the one we want
                    n_if -= 1
                    if index in rng_in_if:
                        unique_rng_in_if_statement[n_if].append(index)
                        break

            #add rng definition to every if condition statement that contains a rng statement 
            for i in range(len(if_statements_indexes)):
                if len(unique_rng_in_if_statement[i]) > 0:
                    begin = begin_indentation[if_statements_indexes[i]] + 1
                    line = body[begin]
                    conditions = conditions_by_line[begin] 
                    # find the number of spaces on the left
                    n_spaces = len(line) - len(line.lstrip(' '))
                    # construct rng object creation
                    #from hashlib import sha1
                    curr_args = ' '*n_spaces + 'curr_args = [' + ', '.join(conditions) + ']'
                    seeder = ' '*n_spaces + "seed = int(sha1(str(curr_args).encode('utf-8')).hexdigest(), 16)"
                    rnger = ' '*n_spaces + 'rng = np.random.default_rng(seed)'
                    body[begin] = [curr_args, seeder, rnger, line]
        else:
            rng_in_if_statement = []
            unique_rng_in_if_statement = []
            rng_in_line = [False for i in range(len_body)]
            for i, line in enumerate(body):
                # find variables in line
                if self.contains_rng(i):
                    rng_in_line[i] = True
            all_rng_indexes = [i for i, x in enumerate(rng_in_line) if x]

        # check for a rng statement in the rest of the script (later and before)
        # any all_rng_indexes that are not in unique_rng_in_if_statement
        rng_in_else = []
        all_in_if_statements = []
        for i in unique_rng_in_if_statement:
            all_in_if_statements += i
        # unique
        all_in_if_statements = np.array(all_in_if_statements, dtype=int)
        all_in_if_statements = np.unique(all_in_if_statements)
        for index in all_rng_indexes: # the unique remaining indexes
            if not index in all_in_if_statements:
                rng_in_else.append(index)
        # where is the for loop of kronbinations
        kronbinations_index = self.kronbinations_index
        # any rng_in_else that are before the kronbinations_index
        rng_in_else_before = []
        rng_in_else_after = []
        for index in rng_in_else:
            if index < kronbinations_index:
                rng_in_else_before.append(index)
            elif index > kronbinations_index:
                rng_in_else_after.append(index)

        # rng definition at beginning of script
        if len(rng_in_else_before) > 0:
            # find the number of spaces on the left
            line = body[0]
            n_spaces = len(line) - len(line.lstrip(' '))
            # construct rng object creation
            rnger = ' '*n_spaces + 'rng = np.random.default_rng('+str(seed)+')'
            body[0] = [rnger, line]

        # after the kronbinations_index, chek if in between if statements -> make a map of lines in if statements
        in_if = np.zeros(len_body, dtype=bool)
        for i in if_statements_indexes:
            begin = begin_indentation[i]
            ender = end_indentation[i]
            in_if[begin:ender] = True
        i = kronbinations_index+1
        how_many = 0
        while i < len_body:
            # check in_if
            if in_if[i]:
                i += 1
            else:
                beginning = i 
                # find next in_if:
                while i < len_body:
                    if in_if[i]:
                        break
                    i += 1
                ending = i
                # check which rng_in_else_after are in this line
                rng_in_interval = []
                for index in rng_in_else_after:
                    if index >= beginning and index < ending:
                        rng_in_interval.append(index)
                if len(rng_in_interval) > 0:
                    how_many += 1
                    # find the number of spaces on the left
                    line = body[beginning]
                    n_spaces = len(line) - len(line.lstrip(' '))
                    # construct rng object creation
                    curr_args = ' '*n_spaces + 'curr_args = [' + ', '.join(for_variables) + ']'
                    seeder = ' '*n_spaces + "seed = int(sha1(str(curr_args).encode('utf-8')).hexdigest(), 16)"
                    rnger = ' '*n_spaces + 'rng = np.random.default_rng(seed)'
                    body[beginning] = [curr_args, seeder, rnger, line]
        # expand list of lists into one list
        new_body = []
        for line in body:
            if isinstance(line, list):
                new_body += line
            else:
                new_body.append(line)
        body = new_body
        #dline()
        #for line in body:
        #    print(line)
        return body

    # a function that constructs a function string from the lines of the definition and the new_body 
    def construct_function_string(self):
        # construct function string
        #self.definition, self.body, self.name, self.input_args,
        import_statements = self.import_statements
        input_args = self.input_args
        # remove rng if found
        if 'rng' in input_args:
            input_args.remove('rng')
        # import numpy as np
        function_string = ''
        if not import_statements is None:
            for raw_statement in import_statements:
                if isinstance(raw_statement, str):
                    if raw_statement.startswith('from') or raw_statement.startswith('import'):
                        statement = raw_statement
                    else:
                        statement = 'from ' + raw_statement + ' import *'
                elif isinstance(raw_statement, list):
                    statement = 'import ' + raw_statement[0] + ' as ' + raw_statement[1]
                else:
                    raise Exception('Unknown import statement: use list for "import x as y" or string for "from x import *" or full import statement "x"')
                function_string += statement + '\n'
            if not 'import numpy as np' in import_statements:
                function_string += 'import numpy as np\n'
        else:
            function_string += 'import numpy as np\n'
        function_string += 'from hashlib import sha1\n\n'
        # other func
        for other_func in self.other_func:
            curr_str = inspect.getsource(other_func)
            function_string += curr_str + '\n\n'

        # construct definition
        function_string += 'def ' + self.name + '(' + ', '.join(input_args) + '):\n'
        #function_string = self.definition + '\n'
        for line in self.new_body:
            function_string += line + '\n'
        return function_string

    def import_functions_from_file(self):
        filename = self.file_name
        function_name = self.name
        spec = importlib.util.spec_from_file_location(filename, filename)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        functions = getattr(module, function_name)
        return functions
# Test
#from kronbinations import *

#def gridspace(k, a, b, rng):
#    d = rng.uniform(0, 1.0)
#    for i, v, c in k.kronprod(index=True, changed=True):
#        if c[1] == c[0]:
#            f = rng.uniform(a, b)
#        A[i] = c[0] + f + d
#        e = rng.uniform(a, b)
#        B[i] = c[1] + f + e
#    return A, B

#FM = Kron_Fun_Modifier(gridspace)
#FM.new_body 

#FM. find_all_variables()
#for i in [3]: #range(len(FM.body)):
#    print(FM.variables_in_line(i))
#    dline()
#FM.kronbination_separation()


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "kronbinations",
    version                    = "1.14",
    author = "Michael Schilling",
    author_email = "michael@ntropic.de",
    description  = "kronbinations is used to remove nested loops and perform parameter sweeps.",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/Ntropic/kronbinations/archive/refs/tags/v1.14.tar.gz",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    install_requires=["numpy", "weighted_tqdm"],
    python_requires=">=3.6",
)



# -*- coding: utf-8 -*-
import numpy as np
import itertools
from weighted_tqdm import weighted_kronbinations_tqdm, weighted_tqdm
from .Kron_Array import *

class kronbinations():
    # A class for scanning parameter landscapes
    # creates iterators over multiple parameters via kronecker products (itertools product), 
    # automatizes the construction of arrays to sotore results on the landscape, simplifies indexing 
    # while keeping the ability to add functions that get only executed in a specific subloop by tracking when a variable is changed

    def __init__(self, *values, **kwargs):
        # If values is a dictionary, also store the keys, so that change can be outputted by key
        weights_given = True if 'weights' in kwargs.keys() else False
        if isinstance(values[0], dict):
            if len(values) > 1:
                raise ValueError('If values is a dictionary, it must be the only argument')
            input_values = values[0]
            keys = input_values.keys()
            values = input_values.values()
            self.array_vars_all_names, self.array_vars_all, self.all_weights = [], [], []
            for i, (key, value) in enumerate(zip(keys, values)):  
                # check if value is a dict with value and weight
                self.array_vars_all_names.append(key)
                if isinstance(value, dict):
                    if 'value' in value.keys():
                        self.array_vars_all.append(value['value'])
                    else:
                        raise ValueError('If value is a dictionary, it must have a value key') 
                    if 'weight' in value.keys():
                        self.all_weights.append(value['weight'])
                    else:
                        self.all_weights.append(None)
                else: # Normal 
                    self.array_vars_all.append(value)
                    if weights_given:
                        self.all_weights.append(kwargs['weights'][i])
                    else:
                        self.all_weights.append(None)
            self.return_as_dict = True
        else:
            self.array_vars_all = list(values)
            if weights_given:
                self.all_weights = list(kwargs['weights'])
            else:
                self.all_weights = [None] * len(self.array_vars_all)
            self.return_as_dict = False
            self.array_vars_all_names = None
        for i, arr in enumerate(self.array_vars_all):
            # if does not have length, transform into array
            if not hasattr(arr, '__len__'):
                self.array_vars_all[i] = np.array([arr])
        # only relevant values in array_directions
        self.array_vars = [arr for arr in self.array_vars_all if len(arr) > 1]
        if isinstance(values, dict):
            self.array_vars_names = [name for name, arr in zip(self.array_vars_all_names, self.array_vars_all) if len(arr) > 1]
        # add index values of the array vars 
        self.array_vars_indexes = [i for i, arr in enumerate(self.array_vars_all) if len(arr) > 1] 
        self.weights = [w for i, w in enumerate(self.all_weights) if i in self.array_vars_indexes]

        if self.return_as_dict:
            self.curr_vals = {key: arr[0] for key, arr in zip(self.array_vars_all_names, self.array_vars_all)}
        else:
            self.curr_vals = [arr[0] for arr in self.array_vars_all]

        self.array_lengths_all = [len(arr) for arr in self.array_vars_all]
        self.array_lengths = [len(arr) for arr in self.array_vars]
        self.shape_all = tuple(self.array_lengths_all)
        self.shape = tuple(self.array_lengths)
        self.ndim_all = len(self.array_lengths_all)
        self.ndim = len(self.array_lengths)
        self.total_length = np.prod(self.array_lengths)
        self.size_all = np.prod(self.array_lengths_all)
        self.size = np.prod(self.array_lengths)
        
        self.index_list = [np.arange(len(v)) for v in self.array_vars]
        self.index_list_all = [np.arange(len(v)) for v in self.array_vars_all]
        # Define the iterators
        self.setup_iterator()

        self.do_index = True
        self.do_change = True
        self.do_tqdm = True
        self.set(**kwargs)   # redo these values if passed as kwargs
        # Initialize tqdm object
        self.pbar = weighted_kronbinations_tqdm(self.array_vars, self.weights, self.size)


    def empty(self, *var, **args):
        return Kron_Array(self.array_lengths, 'empty', *var, **args)
    def ones(self, *var, **args):
        return Kron_Array(self.array_lengths, 'ones', *var, **args)
    def zeros(self, *var, **args):
        return Kron_Array(self.array_lengths, 'zeros', *var, **args)
    def full(self, *var, **args):
        return Kron_Array(self.array_lengths, 'full', *var, **args)
    def random(self, *var, **args):
        return Kron_Array(self.array_lengths, 'random', *var, **args)

    def set(self, **args):
        key_substitution_list = [['index', 'do_index'], ['change', 'do_change'], ['progress', 'do_tqdm']]
        key_list = [v[0] for v in key_substitution_list]
        subs_list = [v[1] for v in key_substitution_list]
        for key, value in args.items():
            # Substitute certain keys from substitution list
            if key in key_list:
                key = subs_list[key_list.index(key)]
            if (key == 'return_as_dict' and value==True) and not isinstance(self.array_vars_all_names, list):
                raise ValueError('Keys are not defined, must create Object via dictionary in order to set "return_as_dict = True".')
            else:
                setattr(self, key, value)
        if self.do_tqdm:
            self.pbar = weighted_kronbinations_tqdm(self.array_vars, self.weights, self.size)
    def get(self, *args):
        key_substitution_list = [['index', 'do_index'], ['change', 'do_change'], ['progress', 'do_tqdm']]
        key_list = [v[0] for v in key_substitution_list]
        subs_list = [v[1] for v in key_substitution_list]
        x = []
        for key in args:
            if key in key_list:
                key = subs_list[key_list.index(key)]
            x.append(getattr(self, key))
        if len(x) == 1:
            return x[0]
        else:
            return x

    def __getitem__(self, key):
        # If the key is not in the data, return None
        if key in key_list:
            key = subs_list[key_list.index(key)]
            print(key)
        if key not in self.data:
            return None
        else:
            return self.data[key]

    def setup_iterator(self):
        self.product = itertools.product(*self.array_vars_all)
        self.indexes = itertools.product(*self.index_list)
        self.indexes_all = itertools.product(*self.index_list_all)

        last_indexes = -np.ones(self.ndim, dtype=int)
        last_indexes_all = -np.ones(self.ndim_all, dtype=int)
        last_values = [v[0] for v in self.array_vars_all]
        changed_var = np.zeros(self.ndim_all, dtype=bool)
        if self.return_as_dict:
            self.last_values = dict(zip(self.array_vars_all_names, last_values))
            self.last_indexes = last_indexes
            self.last_indexes_all = last_indexes_all
            self.changed_var = dict(zip(self.array_vars_all_names, changed_var))
        else:   
            self.last_values = last_values
            self.last_indexes = last_indexes
            self.last_indexes_all = last_indexes_all
            self.changed_var = changed_var

    def __next__(self):
        last_values = next(self.product)
        curr_index = next(self.indexes)
        curr_index_all = next(self.indexes_all)
        changed_var = tuple(np.not_equal(curr_index_all, self.last_indexes_all))
        if self.return_as_dict:
            self.last_values = dict(zip(self.array_vars_all_names, last_values))
            self.last_indexes = curr_index
            self.last_indexes_all = curr_index_all
            self.changed_var = dict(zip(self.array_vars_all_names, changed_var))
        else:   
            self.last_values = last_values
            self.last_indexes = curr_index
            self.last_indexes_all = curr_index_all
            self.changed_var = changed_var
        
        return self.last_values, self.last_indexes, self.changed_var

    def kronprod(self, **args):
        self.set(**args)
        if self.do_tqdm:
            # get all indexes
            self.pbar.init(np.array(list(itertools.product(*self.index_list))))
        if self.do_index:
            if self.do_change:
                for n in range(self.total_length):
                    v,i,c = next(self)
                    yield i, v, c
                    if self.do_tqdm:
                        self.pbar.increment()
            else:
                for n in range(self.total_length):
                    v,i,_ = next(self)
                    yield i, v
                    if self.do_tqdm:
                        self.pbar.increment()
        else:
            if self.do_change: 
                for n in range(self.total_length):
                    v,_,c = next(self)
                    yield v, c
                    if self.do_tqdm:
                        self.pbar.increment()
            else:  
                for n in range(self.total_length):
                    v,_,_ = next(self)
                    yield v
                    if self.do_tqdm:
                        self.pbar.increment()
        if self.do_tqdm:
            self.pbar.close()
        self.setup_iterator()
        
    def tqdm(self, iterator, weights=None, name='', **kwargs):
        if self.do_tqdm:
            return self.pbar.sub_tqdm(iterator, weights=weights, name=name, **kwargs)
        else:
            return weighted_tqdm(iterator, weights=weights, name=name, **kwargs)

    def changed(self, elem=None):
        if elem is None:
            return self.changed_var
        elif isinstance(elem, int):
            if self.return_as_dict:
                string = self.array_vars_all_names[elem]
                return self.changed_var[string]
            else:
                return self.changed_var[elem]
        elif isinstance(elem, str): # Outputs changed by key
            if isinstance(self.array_vars_all_names, list):
                return self.changed_var[elem]
            else:
                raise ValueError('Keys are not defined, must create Object via dictionary for this functionality.')
                
    def index(self, elem=None):
        if elem is None:
            return self.last_indexes
        elif isinstance(elem, int):
            return self.last_indexes[elem]
        elif isinstance(elem, str): # By key
            if isinstance(self.array_vars_all_names, list):
                ind = self.array_vars_all_names.index(elem)
                return self.last_indexes_all[ind]
            else:
                raise ValueError('Keys are not defined, must create Object via dictionary for this functionality.')

    def value(self, elem=None):
        if elem is None:
            return self.last_values
        elif isinstance(elem, int):
            # is dictionary?
            if self.return_as_dict:
                string = self.array_vars_all_names[elem]
                return self.last_values[string]
            else:
                return self.last_values[elem]
        elif isinstance(elem, str):
            if isinstance(self.array_vars_all_names, list):
                return self.last_values[elem]
            else:
                raise ValueError('Keys are not defined, must create Object via dictionary for this functionality.')
    
    def output_definition_kronprod(self, **args):
        self.set(**args)
        i = 'index'
        v = 'value'
        c = 'change'
        if self.do_index:
            if self.do_change:
                return i, v, c
            else:
                return i, v
        else:
            if self.do_change: 
                return v, c
            else:
                return v, 

    def all_combinations_array(self):
        # Generate an array for every combination of the input arrays
        # initialize arrays
        array_vars_all = []
        for i in range(self.ndim_all):
            curr_param = self.array_vars_all[i] # if curr_param doesn't have length, then make it an array
            if not hasattr(curr_param, '__len__'):
                curr_param = np.array([curr_param])
            curr_arr = np.empty(self.array_lengths_all, dtype=curr_param.dtype)
            # loop over all combinations
            for j in range(len(curr_param)):
                curr_arr[(slice(None), ) * i + (j,) ] = curr_param[j]
            array_vars_all.append(curr_arr)
        return array_vars_all
    
    